# https://www.mp3-tech.org/programmer/frame_header.html
# http://gabriel.mp3-tech.org/mp3infotag.html

__all__ = [
	'LAMEEncodingFlags',
	'LAMEHeader',
	'LAMEReplayGain',
	'MP3',
	'MP3StreamInfo',
	'MPEGFrameHeader',
	'VBRIHeader',
	'VBRIToC',
	'XingHeader',
	'XingToC',
]

import os
import re
import struct
from functools import lru_cache

import more_itertools
from attr import (
	attrib,
	attrs,
)
from tbm_utils import (
	AttrMapping,
	LabelList,
	humanize_filesize,
)

from .id3v1 import ID3v1
from .id3v2 import (
	ID3v2,
	ID3v2Frames,
)
from .tables import (
	LAMEBitrateMode,
	LAMEChannelMode,
	LAMEPreset,
	LAMEReplayGainOrigin,
	LAMEReplayGainType,
	LAMESurroundInfo,
	MP3BitrateMode,
	MP3Bitrates,
	MP3ChannelMode,
	MP3SampleRates,
	MP3SamplesPerFrame,
)
from ..exceptions import (
	InvalidFormat,
	InvalidFrame,
	InvalidHeader,
)
from ..models import (
	Format,
	StreamInfo,
)
from ..utils import (
	datareader,
	humanize_bitrate,
	humanize_sample_rate,
)

try:  # pragma: nocover
	import bitstruct.c as bitstruct
	bitstruct.Error = (TypeError, ValueError)
except ImportError:  # pragma: nocover
	import bitstruct
	bitstruct.Error = (bitstruct.Error,)


@attrs(repr=False)
class LAMEReplayGain(AttrMapping):
	peak = attrib()
	track_type = attrib(converter=LAMEReplayGainType)
	track_origin = attrib(converter=LAMEReplayGainOrigin)
	track_adjustment = attrib()
	album_type = attrib(converter=LAMEReplayGainType)
	album_origin = attrib(converter=LAMEReplayGainOrigin)
	album_adjustment = attrib()

	@datareader
	@classmethod
	def load(cls, data):
		peak_data = struct.unpack('>I', data.read(4))[0]

		if peak_data == 0:
			gain_peak = None
		else:
			gain_peak = peak_data / 2 ** 23

		track_gain_type_, track_gain_origin_, track_gain_sign, track_gain_adjustment_ = bitstruct.unpack(
			'u3 u3 b1 u9',
			data.read(2),
		)

		track_gain_type = LAMEReplayGainType(track_gain_type_)
		track_gain_origin = LAMEReplayGainOrigin(track_gain_origin_)
		track_gain_adjustment = track_gain_adjustment_ / 10.0

		if track_gain_sign:
			track_gain_adjustment *= -1

		album_gain_type_, album_gain_origin_, album_gain_sign, album_gain_adjustment_ = bitstruct.unpack(
			'u3 u3 b1 u9',
			data.read(2),
		)

		album_gain_type = LAMEReplayGainType(album_gain_type_)
		album_gain_origin = LAMEReplayGainOrigin(album_gain_origin_)
		album_gain_adjustment = album_gain_adjustment_ / 10.0

		if album_gain_sign:
			album_gain_adjustment *= -1

		return cls(
			gain_peak,
			track_gain_type,
			track_gain_origin,
			track_gain_adjustment,
			album_gain_type,
			album_gain_origin,
			album_gain_adjustment,
		)


@attrs(repr=False)
class LAMEEncodingFlags(AttrMapping):
	nogap_continuation = attrib(converter=bool)
	nogap_continued = attrib(converter=bool)
	nssafejoint = attrib(converter=bool)
	nspsytune = attrib(converter=bool)


@attrs(repr=False)
class LAMEHeader(AttrMapping):
	_crc = attrib()
	version = attrib()
	revision = attrib()
	ath_type = attrib()
	audio_crc = attrib()
	audio_size = attrib()
	bitrate = attrib()
	bitrate_mode = attrib(converter=LAMEBitrateMode)
	channel_mode = attrib(converter=LAMEChannelMode)
	delay = attrib()
	encoding_flags = attrib(converter=LAMEEncodingFlags.from_mapping)
	lowpass_filter = attrib()
	mp3_gain = attrib()
	noise_shaping = attrib()
	padding = attrib()
	preset = attrib(converter=LAMEPreset)
	replay_gain = attrib()
	source_sample_rate = attrib()
	surround_info = attrib(converter=LAMESurroundInfo)
	unwise_settings_used = attrib(converter=bool)

	def __repr__(self):
		repr_dict = {}

		for k, v in sorted(self.items()):
			if k == 'bitrate':
				repr_dict[k] = humanize_bitrate(v)
			elif k == 'audio_size':
				repr_dict[k] = humanize_filesize(v, precision=2)
			elif not k.startswith('_'):
				repr_dict[k] = v

		return super().__repr__(repr_dict=repr_dict)

	@datareader
	@classmethod
	def load(cls, data, xing_quality):
		encoder = data.read(9)
		if not encoder.startswith(b'LAME'):
			raise InvalidHeader('Valid LAME header not found.')

		version = None
		version_match = re.search(rb'LAME(\d+)\.(\d+)', encoder)
		if version_match:  # pragma: nobranch
			version = tuple(int(part) for part in version_match.groups())

		revision, bitrate_mode_ = bitstruct.unpack(
			'u4 u4',
			data.read(1),
		)
		bitrate_mode = LAMEBitrateMode(bitrate_mode_)

		# TODO: Decide what, if anything, to do with the different meanings in LAME.
		# quality = (100 - xing_quality) % 10
		# vbr_quality = (100 - xing_quality) // 10

		lowpass_filter = struct.unpack(
			'B',
			data.read(1),
		)[0] * 100

		replay_gain = LAMEReplayGain.load(data)

		flags_ath = bitstruct.unpack_dict(
			'b1 b1 b1 b1 u4',
			[
				'nogap_continuation',
				'nogap_continued',
				'nssafejoint',
				'nspsytune',
				'ath_type',
			],
			data.read(1),
		)

		ath_type = flags_ath.pop('ath_type')
		encoding_flags = LAMEEncodingFlags(
			flags_ath['nogap_continuation'],
			flags_ath['nogap_continued'],
			flags_ath['nssafejoint'],
			flags_ath['nspsytune'],
		)

		# TODO: Different representation for VBR minimum bitrate vs CBR/ABR specified bitrate?
		# Can only go up to 255.
		bitrate = struct.unpack(
			'B',
			data.read(1),
		)[0] * 1000

		delay, padding = bitstruct.unpack(
			'u12 u12',
			data.read(3),
		)

		source_sample_rate, unwise_settings_used, channel_mode_, noise_shaping = bitstruct.unpack(
			'u2 b1 u3 u2',
			data.read(1),
		)
		channel_mode = LAMEChannelMode(channel_mode_)

		mp3_gain = bitstruct.unpack(
			's8',
			data.read(1),
		)[0]

		surround_info_, preset_used_ = bitstruct.unpack(
			'p2 u3 u11',
			data.read(2),
		)
		surround_info = LAMESurroundInfo(surround_info_)

		preset = LAMEPreset(preset_used_)

		audio_size, audio_crc, lame_crc = struct.unpack(
			'>I2s2s',
			data.read(8),
		)

		return cls(
			lame_crc,
			version,
			revision,
			ath_type,
			audio_crc,
			audio_size,
			bitrate,
			bitrate_mode,
			channel_mode,
			delay,
			encoding_flags,
			lowpass_filter,
			mp3_gain,
			noise_shaping,
			padding,
			preset,
			replay_gain,
			source_sample_rate,
			surround_info,
			unwise_settings_used,
		)


class XingToC(LabelList):
	item_label = 'entries'


@attrs(repr=False)
class XingHeader(AttrMapping):
	_lame = attrib()
	num_frames = attrib()
	num_bytes = attrib()
	toc = attrib(converter=XingToC)
	quality = attrib()

	@datareader
	@classmethod
	def load(cls, data):
		if data.read(4) not in [b'Xing', b'Info']:
			raise InvalidHeader('Valid Xing header not found.')

		flags = struct.unpack('>i', data.read(4))[0]

		num_frames = num_bytes = toc = quality = lame_header = None

		if flags & 1:  # pragma: nobranch
			num_frames = struct.unpack('>I', data.read(4))[0]

		if flags & 2:  # pragma: nobranch
			num_bytes = struct.unpack('>I', data.read(4))[0]

		if flags & 4:  # pragma: nobranch
			toc = XingToC(bytearray(data.read(100)))

		if flags & 8:  # pragma: nobranch
			quality = struct.unpack('>I', data.read(4))[0]

		if data.peek(4) == b'LAME':
			lame_header = LAMEHeader.load(data, quality)

		return cls(lame_header, num_frames, num_bytes, toc, quality)


class VBRIToC(LabelList):
	item_label = 'entries'


@attrs(repr=False)
class VBRIHeader(AttrMapping):
	version = attrib()
	delay = attrib()
	quality = attrib()
	num_bytes = attrib()
	num_frames = attrib()
	num_toc_entries = attrib()
	toc_scale_factor = attrib()
	toc_entry_num_bytes = attrib()
	toc_entry_num_frames = attrib()
	toc = attrib(converter=VBRIToC)

	@datareader
	@classmethod
	def load(cls, data):
		if data.read(4) not in [b'VBRI']:
			raise InvalidHeader('Valid VBRI header not found.')

		version = struct.unpack('>H', data.read(2))[0]
		delay = struct.unpack('>e', data.read(2))[0]
		quality = struct.unpack('>H', data.read(2))[0]
		num_bytes = struct.unpack('>I', data.read(4))[0]
		num_frames = struct.unpack('>I', data.read(4))[0]
		num_toc_entries = struct.unpack('>H', data.read(2))[0]
		toc_scale_factor = struct.unpack('>H', data.read(2))[0]
		toc_entry_num_bytes = struct.unpack('>H', data.read(2))[0]
		toc_entry_num_frames = struct.unpack('>H', data.read(2))[0]

		if toc_entry_num_bytes not in [2, 4]:
			raise InvalidHeader('Invalid VBRI TOC entry size.')

		if toc_entry_num_bytes == 2:
			pattern = '>H'
		else:
			pattern = '>I'

		toc = VBRIToC(
			struct.unpack(pattern, data.read(toc_entry_num_bytes))[0]
			for _ in range(num_toc_entries)
		)

		return cls(
			version,
			delay,
			quality,
			num_bytes,
			num_frames,
			num_toc_entries,
			toc_scale_factor,
			toc_entry_num_bytes,
			toc_entry_num_frames,
			toc,
		)


@attrs(repr=False)
class MPEGFrameHeader(AttrMapping):
	_start = attrib()
	_size = attrib()
	_vbri = attrib()
	_xing = attrib()
	version = attrib()
	layer = attrib()
	protected = attrib()
	padded = attrib(converter=bool)
	bitrate = attrib()
	channel_mode = attrib(converter=MP3ChannelMode)
	channels = attrib()
	sample_rate = attrib()

	def __repr__(self):
		repr_dict = {}

		for k, v in sorted(self.items()):
			if k == 'bitrate':
				repr_dict[k] = humanize_bitrate(v)
			elif k == 'sample_rate':
				repr_dict[k] = humanize_sample_rate(v)
			elif not k.startswith('_'):
				repr_dict[k] = v

		return super().__repr__(repr_dict=repr_dict)

	@datareader
	@classmethod
	def load(cls, data):
		frame_start = data.tell()

		sync, version_id, layer_index, protection = bitstruct.unpack(
			'u11 u2 u2 b1',
			data.read(2),
		)

		if sync != 2047:
			raise InvalidFrame('Not a valid MPEG audio frame.')

		version = [2.5, None, 2, 1][version_id]

		layer = 4 - layer_index

		protected = not protection

		bitrate_index, sample_rate_index, padded = bitstruct.unpack(
			'u4 u2 b1',
			data.read(1),
		)

		if (
			version_id == 1
			or layer_index == 0
			or bitrate_index == 0
			or bitrate_index == 15
			or sample_rate_index == 3
		):
			raise InvalidFrame('Not a valid MPEG audio frame.')

		channel_mode = MP3ChannelMode(bitstruct.unpack('u2', data.read(1))[0])
		channels = 1 if channel_mode == 3 else 2

		bitrate = MP3Bitrates[(version, layer)][bitrate_index] * 1000
		sample_rate = MP3SampleRates[version][sample_rate_index]

		samples_per_frame, slot_size = MP3SamplesPerFrame[(version, layer)]

		frame_size = (((samples_per_frame // 8 * bitrate) // sample_rate) + padded) * slot_size

		vbri_header = None
		xing_header = None
		if layer == 3:  # pragma: nobranch
			if version == 1:
				if channel_mode != 3:
					xing_header_start = 36
				else:
					xing_header_start = 21
			elif channel_mode != 3:
				xing_header_start = 21
			else:
				xing_header_start = 13

			data.seek(frame_start + xing_header_start, os.SEEK_SET)

			if data.peek(4) in [b'Xing', b'Info']:
				xing_header = XingHeader.load(data.read(frame_size))

			data.seek(frame_start + 36, os.SEEK_SET)

			if data.peek(4) == b'VBRI':
				vbri_header = VBRIHeader.load(data)

		return cls(
			frame_start,
			frame_size,
			vbri_header,
			xing_header,
			version,
			layer,
			protected,
			padded,
			bitrate,
			channel_mode,
			channels,
			sample_rate,
		)


@attrs(repr=False)
class MP3StreamInfo(StreamInfo):
	_start = attrib()
	_end = attrib()
	_size = attrib()
	_vbri = attrib()
	_xing = attrib()
	version = attrib()
	layer = attrib()
	protected = attrib()
	bitrate = attrib()
	bitrate_mode = attrib(converter=MP3BitrateMode)
	channel_mode = attrib(converter=MP3ChannelMode)
	channels = attrib()
	duration = attrib()
	sample_rate = attrib()

	@staticmethod
	def count_mpeg_frames(data):
		num_frames = 0

		buffer_size = 128
		buffer = data.peek(buffer_size)

		while len(buffer) >= buffer_size:
			sync_start = buffer.find(b'\xFF')

			if sync_start >= 0:  # pragma: nobranch
				data.seek(sync_start, os.SEEK_CUR)

				try:
					frame = MPEGFrameHeader.load(data)
					num_frames += 1
					data.seek(frame._start + frame._size, os.SEEK_SET)
				except (InvalidFrame, *bitstruct.Error):  # pragma: nocover
					data.seek(1, os.SEEK_CUR)
			else:
				data.seek(buffer_size, os.SEEK_CUR)

			buffer = data.peek(buffer_size)

		return num_frames

	@staticmethod
	@lru_cache()
	def find_mpeg_frames(data):
		frames = []
		cached_frames = None
		buffer_size = 128
		buffer = data.peek(buffer_size)

		while len(buffer) >= buffer_size:
			sync_start = buffer.find(b'\xFF')

			if sync_start >= 0:
				data.seek(sync_start, os.SEEK_CUR)

				if bitstruct.unpack('u11', data.peek(2))[0] == 2047:
					for _ in range(4):
						try:
							frame = MPEGFrameHeader.load(data)
							frames.append(frame)
							if frame._xing and frame._xing.num_frames:
								break
							data.seek(frame._start + frame._size, os.SEEK_SET)
						except (InvalidFrame, *bitstruct.Error):
							data.seek(1, os.SEEK_CUR)
							break
				else:
					data.seek(sync_start + 1, os.SEEK_CUR)

				if frames and (len(frames) >= 4 or frames[0]._xing):
					break

				if len(frames) >= 2 and cached_frames is None:
					cached_frames = frames.copy()

				del frames[:]
			else:
				data.seek(buffer_size, os.SEEK_CUR)

			buffer = data.peek(buffer_size)

		# I actually found a PNG file that had multiple consecutive MPEG frames parsed.
		# The all_equal check combats this false positive by
		# making sure certain attributes don't change between frames.
		if not frames:
			if (
				cached_frames
				and more_itertools.all_equal(
					[
						frame.channel_mode,
						frame.channels,
						frame.layer,
						frame.sample_rate,
						frame.version,
					]
					for frame in cached_frames
				)
			):
				frames = cached_frames
			else:
				raise InvalidFormat("No XING header and insufficient MPEG frames.")

		return frames

	@datareader
	@classmethod
	def load(cls, data):
		frames = cls.find_mpeg_frames(data)

		samples_per_frame, _ = MP3SamplesPerFrame[(frames[0].version, frames[0].layer)]

		data.seek(0, os.SEEK_END)
		end_pos = data.tell()

		# This is an arbitrary amount that should hopefully encompass all end tags.
		# Starting low so as not to add unnecessary processing time.
		chunk_size = 64 * 1024
		if end_pos > chunk_size:
			data.seek(-chunk_size, os.SEEK_END)
		else:
			data.seek(0, os.SEEK_SET)

		end_buffer = data.read()

		end_tag_offset = 0
		for tag_type in [b'APETAGEX', b'LYRICSBEGIN', b'TAG']:
			tag_offset = end_buffer.rfind(tag_type)

			if tag_offset > 0:
				tag_offset = len(end_buffer) - tag_offset

				if tag_offset > end_tag_offset:
					end_tag_offset = tag_offset

		audio_start = frames[0]._start
		audio_end = end_pos - end_tag_offset
		audio_size = audio_end - audio_start

		bitrate_mode = MP3BitrateMode.UNKNOWN

		vbri_header = frames[0]._vbri
		xing_header = frames[0]._xing
		if xing_header:
			if xing_header.num_frames:
				num_samples = samples_per_frame * xing_header.num_frames
			else:
				# Some XING headers have num_frames==0.
				# Manually count all the MPEG frames for bitrate/duration calculations.
				# Remove XING frame from frame count for bitrate calculation accuracy.
				data.seek(frames[0]._start, os.SEEK_SET)
				num_samples = samples_per_frame * (cls.count_mpeg_frames(data) - 1)

			# I prefer to include the Xing/LAME header as part of the audio.
			# Google Music seems to do so for calculating client ID.
			# Haven't tested in too many other scenarios.
			# But, there should be enough low-level info for people to calculate this if desired.
			if xing_header._lame:
				# Old versions of LAME wrote invalid delay/padding
				# for short MP3s with low bitrate.
				# Subtract them only them if there would be samples left.
				lame_padding = xing_header._lame.delay + xing_header._lame.padding
				if lame_padding < num_samples:
					num_samples -= lame_padding

				if xing_header._lame.bitrate_mode in [1, 8]:
					bitrate_mode = MP3BitrateMode.CBR
				elif xing_header._lame.bitrate_mode in [2, 9]:
					bitrate_mode = MP3BitrateMode.ABR
				elif xing_header._lame.bitrate_mode in [3, 4, 5, 6]:
					bitrate_mode = MP3BitrateMode.VBR
		elif vbri_header:
			num_samples = samples_per_frame * vbri_header.num_frames
			bitrate_mode = MP3BitrateMode.VBR
		else:
			if more_itertools.all_equal([frame['bitrate'] for frame in frames]):  # pragma: nobranch
				bitrate_mode = MP3BitrateMode.CBR

			num_samples = samples_per_frame * (audio_size / frames[0]._size)

		if bitrate_mode is MP3BitrateMode.CBR:
			bitrate = frames[0].bitrate
		else:
			# Subtract Xing/LAME frame size from audio_size for bitrate calculation accuracy.
			if xing_header:
				bitrate = ((audio_size - frames[0]._size) * 8 * frames[0].sample_rate) / num_samples
			else:
				bitrate = (audio_size * 8 * frames[0].sample_rate) / num_samples

		duration = (audio_size * 8) / bitrate

		version = frames[0].version
		layer = frames[0].layer
		protected = frames[0].protected
		sample_rate = frames[0].sample_rate
		channel_mode = frames[0].channel_mode
		channels = frames[0].channels

		return cls(
			audio_start,
			audio_end,
			audio_size,
			vbri_header,
			xing_header,
			version,
			layer,
			protected,
			bitrate,
			bitrate_mode,
			channel_mode,
			channels,
			duration,
			sample_rate,
		)


class MP3(Format):
	"""MP3 file format object.

	Extends :class:`Format`.

	Attributes:
		pictures (list): A list of :class:`ID3v2Picture` objects.
		streaminfo (MP3StreamInfo): The audio stream information.
		tags (ID3v2Frames): The ID3v2 tag frames, if present.
	"""

	tags_type = ID3v2Frames

	@classmethod
	def load(cls, data):
		self = super()._load(data)

		try:
			self._id3 = ID3v2.load(self._obj)
			self.pictures = self._id3.pictures
			self.tags = self._id3.tags
		except (InvalidFrame, InvalidHeader):
			self._obj.seek(0, os.SEEK_SET)

		self.streaminfo = MP3StreamInfo.load(self._obj)

		# Use ID3v1 if present and ID3v2 is not.
		if '_id3' not in self:
			self._obj.seek(self.streaminfo._start + self.streaminfo._size, os.SEEK_SET)

			end_buffer = self._obj.read()

			apev2_index = end_buffer.find(b'APETAGEX')
			if apev2_index != -1:
				end_buffer = end_buffer[apev2_index + 8:]

			id3v1_index = end_buffer.find(b'TAG')
			if id3v1_index != -1:
				id3v1 = ID3v1.load(end_buffer[id3v1_index : id3v1_index + 128])
				self._id3 = id3v1
				self.tags = self._id3.tags

		self._obj.close()

		return self



__all__ = [
	'AudioMetadataException',
	'InvalidBlock',
	'InvalidChunk',
	'InvalidComment',
	'InvalidFormat',
	'InvalidFrame',
	'InvalidHeader',
	'UnsupportedFormat',
]


class AudioMetadataException(Exception):
	"""Base exception class."""

	pass


class InvalidBlock(AudioMetadataException):
	"""Exception raised when a FLAC metadata block is invalid."""

	pass


class InvalidComment(AudioMetadataException):
	"""Exception raised when a Vorbis comment is invalid."""

	pass


class InvalidChunk(AudioMetadataException):
	"""Exception raised when a WAV chunk is invalid."""

	pass


class InvalidFormat(AudioMetadataException):
	"""Exception raised when a file format is invalid."""

	pass


class InvalidFrame(AudioMetadataException):
	"""Exception raised when an ID3v2 frame is invalid."""

	pass


class InvalidHeader(AudioMetadataException):
	"""Exception raised when a metadata header is invalid."""

	pass


class UnsupportedFormat(AudioMetadataException):
	"""Exception raised when loading a file that isn't supported."""

	pass



# http://id3.org/ID3v1

__all__ = [
	'ID3v1',
	'ID3v1Fields',
]

from tbm_utils import AttrMapping

from .tables import ID3v1Genres
from ..exceptions import InvalidHeader
from ..models import Tags
from ..utils import datareader


class ID3v1Fields(Tags):
	@datareader
	@classmethod
	def load(cls, data):
		self = cls()

		title = data.read(30).strip(b'\x00').decode('iso-8859-1')
		artist = data.read(30).strip(b'\x00').decode('iso-8859-1')
		album = data.read(30).strip(b'\x00').decode('iso-8859-1')
		year = data.read(4).strip(b'\x00').decode('iso-8859-1')
		comment = data.read(29).strip(b'\x00').decode('iso-8859-1')
		tracknumber = str(data.read(1)[0])
		genre_index = int.from_bytes(data.read(1), byteorder='big')

		if title:
			self.title = [title]

		if artist:
			self.artist = [artist]

		if album:
			self.album = [album]

		if year:
			self.year = [year]

		if comment:
			self.comment = [comment]

		if tracknumber != '0':
			self.tracknumber = [tracknumber]

		if genre_index < len(ID3v1Genres):
			self.genre = [ID3v1Genres[genre_index]]

		return self


class ID3v1(AttrMapping):
	@datareader
	@classmethod
	def load(cls, data):
		if data.read(3) != b"TAG":
			raise InvalidHeader("Valid ID3v1 header not found.")

		self = cls()
		self.tags = ID3v1Fields.load(data)

		return self



# https://xiph.org/vorbis/doc/v-comment.html

__all__ = [
	'VorbisComment',
	'VorbisComments',
]

import struct
from collections import defaultdict

from attr import attrib, attrs
from tbm_utils import AttrMapping

from ..exceptions import InvalidComment
from ..models import Tags
from ..utils import datareader


@attrs(repr=False)
class VorbisComment(AttrMapping):
	name = attrib(converter=lambda n: n.lower())
	value = attrib()

	@datareader
	@classmethod
	def load(cls, data):
		length = struct.unpack('I', data.read(4))[0]
		comment = data.read(length).decode('utf-8', 'replace')

		if '=' not in comment:
			raise InvalidComment("Vorbis comment must contain an '='.")

		name, value = comment.split('=', 1)

		return cls(name, value)


# TODO: Number frames.
class VorbisComments(Tags):
	@datareader
	@classmethod
	def load(cls, data):
		vendor_length = struct.unpack('I', data.read(4))[0]
		vendor = data.read(vendor_length).decode('utf-8', 'replace')
		num_comments = struct.unpack('I', data.read(4))[0]

		fields = defaultdict(list)

		for _ in range(num_comments):
			comment = VorbisComment.load(data)
			fields[comment.name].append(comment.value)

		return cls(**fields, _vendor=vendor)



import sys
import warnings

from .__about__ import *
from .api import *
from .exceptions import *
from .formats import *
from .models import *

__all__ = [
	*__about__.__all__,
	*api.__all__,
	*exceptions.__all__,
	*formats.__all__,
	*models.__all__,
]


# Override warning output format.
def showwarning(message, category, filename, lineno, file=None, line=None):
	if file is None:
		file = sys.stderr
		if file is None:
			return

	delim = '\n    '
	nl = '\n'
	s = f"{category.__name__}:{delim}{delim.join(line for line in str(message).split(nl))}\n"

	try:
		file.write(s)
	except (IOError, UnicodeError):
		pass


warnings.showwarning = showwarning
del showwarning



# https://xiph.org/flac/format.html

__all__ = [
	'FLAC',
	'FLACApplication',
	'FLACCueSheet',
	'FLACCueSheetIndex',
	'FLACCueSheetTrack',
	'FLACMetadataBlock',
	'FLACPadding',
	'FLACPicture',
	'FLACSeekPoint',
	'FLACSeekTable',
	'FLACStreamInfo',
]

import binascii
import struct

from attr import (
	Factory,
	attrib,
	attrs,
)
from tbm_utils import (
	AttrMapping,
	LabelList,
)

from .id3v2 import ID3v2
from .tables import (
	FLACMetadataBlockType,
	ID3PictureType,
)
from .vorbis import VorbisComments
from ..exceptions import (
	InvalidBlock,
	InvalidHeader,
)
from ..models import (
	Format,
	Picture,
	StreamInfo,
)
from ..utils import datareader

try:  # pragma: nocover
	import bitstruct.c as bitstruct
except ImportError:  # pragma: nocover
	import bitstruct


@attrs(repr=False)
class FLACApplication(AttrMapping):
	"""Application metadata block.

	Attributes:
		id (str): The 32-bit application identifier.
		data (bytes): The data defined by the application.
	"""

	id = attrib()  # noqa
	data = attrib()

	def __repr__(self):
		return f"<FLACApplication ({self.id})>"

	@datareader
	@classmethod
	def load(cls, data):
		id_ = data.read(4).decode('utf-8', 'replace')
		data = data.read()

		return cls(id_, data)


@attrs(repr=False)
class FLACCueSheetIndex(AttrMapping):
	"""A cue sheet track index point.

	Attributes:
		number (int): The index point number.

			The first index in a track must have a number of 0 or 1.

			Index numbers must increase by 1 and be unique within a track.

			For CD-DA, an index number of 0 corresponds to the track pre-gab.

		offset (int): Offset in samples relative to the track offset.
	"""

	number = attrib()
	offset = attrib()

	@datareader
	@classmethod
	def load(cls, data):
		offset = struct.unpack(
			'>Q',
			data.read(8),
		)[0]

		number = struct.unpack(
			'>B',
			data.read(1),
		)[0]

		data.read(3)

		return cls(number, offset)


@attrs(repr=False)
class FLACCueSheetTrack(AttrMapping):
	"""A FLAC cue sheet track.

	Attributes:
		track_number (int): The track number of the track.

			0 is not allowed to avoid conflicting with the CD-DA spec lead-in track.

			For CD-DA, the track number must be 1-99 or 170 for the lead-out track.

			For non-CD-DA, the track number must be 255 for the lead-out track.

			Track numbers must be unique within a cue sheet.
		offset (int): Offset in samples relative to the beginning of the FLAC audio stream.
		isrc (str): The ISRC (International Standard Recording Code) of the track.
		type (int): ``0`` for audio, ``1`` for non-audio.
		pre_emphasis (bool): ``True`` if contains pre-emphasis, ``False`` if not.
		indexes (list): The index points for the track as :class:`FLACCueSheetIndex` objects.
	"""

	track_number = attrib()
	offset = attrib()
	isrc = attrib()
	type = attrib()  # noqa
	pre_emphasis = attrib()
	indexes = attrib(default=Factory(list))

	@datareader
	@classmethod
	def load(cls, data):
		offset = struct.unpack(
			'>Q',
			data.read(8),
		)[0]
		track_number = struct.unpack(
			'>B',
			data.read(1),
		)[0]
		isrc = data.read(12).rstrip(b'\x00').decode('ascii', 'replace')

		type_, pre_emphasis = bitstruct.unpack(
			'u1 b1',
			data.read(1),
		)

		data.read(13)
		num_indexes = struct.unpack(
			'>B',
			data.read(1),
		)[0]

		indexes = []
		for _ in range(num_indexes):
			indexes.append(FLACCueSheetIndex.load(data))

		return cls(
			track_number,
			offset,
			isrc,
			type_,
			pre_emphasis,
			indexes,
		)


class FLACCueSheet(LabelList):
	"""The cue sheet metadata block.

	A list-like structure of :class:`FLACCueSheetTrack` objects
	along with some information used in the cue sheet.

	Attributes:
		catalog_number (str): The media catalog number.
		lead_in_samples (int): The number of lead-in samples.
			This is only meaningful for CD-DA cuesheets.
			For others, it should be 0.
		compact_disc (bool): ``True`` if the cue sheet corresponds to a compact disc, else ``False``.
	"""

	item_label = 'tracks'

	def __init__(self, tracks, catalog_number, lead_in_samples, compact_disc):
		super().__init__(tracks)
		self.catalog_number = catalog_number
		self.lead_in_samples = lead_in_samples
		self.compact_disc = compact_disc

	@datareader
	@classmethod
	def load(cls, data):
		catalog_number = data.read(128).rstrip(b'\0').decode('ascii', 'replace')
		lead_in_samples = struct.unpack(
			'>Q',
			data.read(8),
		)[0]
		compact_disc = bitstruct.unpack(
			'b1',
			data.read(1),
		)[0]

		data.read(258)
		num_tracks = struct.unpack(
			'B',
			data.read(1),
		)[0]

		tracks = []
		for _ in range(num_tracks):
			tracks.append(FLACCueSheetTrack.load(data))

		return cls(
			tracks,
			catalog_number,
			lead_in_samples,
			compact_disc,
		)


@attrs(repr=False)
class FLACMetadataBlock(AttrMapping):
	type = attrib()  # noqa
	data = attrib()

	def __repr__(self):
		return f"<FLACMetadataBlock [{self.type}] ({len(self.data)} bytes)>"


@attrs(repr=False)
class FLACPadding(AttrMapping):
	size = attrib()

	def __repr__(self):
		return f"<FLACPadding ({self.size} bytes)>"

	@datareader
	@classmethod
	def load(cls, data):
		return cls(len(data.peek()))


class FLACPicture(Picture):
	@datareader
	@classmethod
	def load(cls, data):
		type_, mime_length = struct.unpack('>2I', data.read(8))
		mime_type = data.read(mime_length).decode('utf-8', 'replace')

		desc_length = struct.unpack('>I', data.read(4))[0]
		description = data.read(desc_length).decode('utf-8', 'replace')

		width, height, depth, colors = struct.unpack('>4I', data.read(16))

		data_length = struct.unpack('>I', data.read(4))[0]
		data = data.read(data_length)

		return cls(
			type=ID3PictureType(type_),
			mime_type=mime_type,
			description=description,
			width=width,
			height=height,
			depth=depth,
			colors=colors,
			data=data,
		)


@attrs(repr=False)
class FLACSeekPoint(AttrMapping):
	first_sample = attrib()
	offset = attrib()
	num_samples = attrib()

	@datareader
	@classmethod
	def load(cls, data):
		return cls(*struct.unpack('>QQH', data.read()))


class FLACSeekTable(LabelList):
	item_label = 'seekpoints'

	@datareader
	@classmethod
	def load(cls, data):
		seekpoints = []
		seekpoint = data.read(18)
		while len(seekpoint) == 18:
			seekpoints.append(FLACSeekPoint.load(seekpoint))
			seekpoint = data.read(18)

		return cls(seekpoints)


@attrs(repr=False)
class FLACStreamInfo(StreamInfo):
	_start = attrib()
	_size = attrib()
	_min_block_size = attrib()
	_max_block_size = attrib()
	_min_frame_size = attrib()
	_max_frame_size = attrib()
	bit_depth = attrib()
	bitrate = attrib()
	channels = attrib()
	duration = attrib()
	md5 = attrib()
	sample_rate = attrib()

	@datareader
	@classmethod
	def load(cls, data):
		stream_info_block_data = bitstruct.unpack(
			'u16 u16 u24 u24 u20 u3 u5 u36 r128',
			data.read(34),
		)

		min_block_size = stream_info_block_data[0]
		max_block_size = stream_info_block_data[1]
		min_frame_size = stream_info_block_data[2]
		max_frame_size = stream_info_block_data[3]
		sample_rate = stream_info_block_data[4]
		channels = stream_info_block_data[5] + 1
		bit_depth = stream_info_block_data[6] + 1
		total_samples = stream_info_block_data[7]
		md5sum = binascii.hexlify(stream_info_block_data[8]).decode('ascii', 'replace')
		duration = total_samples / sample_rate

		return cls(
			None,
			None,
			min_block_size,
			max_block_size,
			min_frame_size,
			max_frame_size,
			bit_depth,
			None,
			channels,
			duration,
			md5sum,
			sample_rate,
		)


class FLAC(Format):
	"""FLAC file format object.

	Extends :class:`Format`.

	Attributes:
		cuesheet (FLACCueSheet): The cuesheet metadata block.
		pictures (list): A list of :class:`FLACPicture` objects.
		seektable (FLACSeekTable): The seektable metadata block.
		streaminfo (FLACStreamInfo): The audio stream information.
		tags (VorbisComments): The Vorbis comment metadata block.
	"""

	tags_type = VorbisComments

	def __init__(self):
		super().__init__()
		self._blocks = []

	@classmethod
	def load(cls, data):
		self = super()._load(data)

		# Ignore ID3v2 in FLAC.
		if self._obj.peek(3) == b'ID3':
			ID3v2.load(self._obj)

		if self._obj.read(4) != b'fLaC':
			raise InvalidHeader("Valid FLAC header not found.")

		header_data = self._obj.read(4)

		while True:
			is_last_block, block_type, block_size = bitstruct.unpack(
				'b1 u7 u24',
				header_data,
			)

			# There are examples of tools writing incorrect block sizes.
			# The FLAC reference implementation unintentionally (I hope?) parses them.
			# I've chosen not to add special handling for these invalid files.
			# If needed, mutagen (https://github.com/quodlibet/mutagen) may support them.
			metadata_block_data = self._obj.read(block_size)

			if block_type == FLACMetadataBlockType.STREAMINFO:
				streaminfo_block = FLACStreamInfo.load(metadata_block_data)
				self.streaminfo = streaminfo_block
				self._blocks.append(streaminfo_block)
			elif block_type == FLACMetadataBlockType.PADDING:
				self._blocks.append(FLACPadding.load(metadata_block_data))
			elif block_type == FLACMetadataBlockType.APPLICATION:
				application_block = FLACApplication.load(metadata_block_data)
				self._blocks.append(application_block)
			elif block_type == FLACMetadataBlockType.SEEKTABLE:
				seektable = FLACSeekTable.load(metadata_block_data)
				self.seektable = seektable
				self._blocks.append(seektable)
			elif block_type == FLACMetadataBlockType.VORBIS_COMMENT:
				comment_block = VorbisComments.load(metadata_block_data)
				self.tags = comment_block
				self._blocks.append(comment_block)
			elif block_type == FLACMetadataBlockType.CUESHEET:
				cuesheet_block = FLACCueSheet.load(metadata_block_data)
				self.cuesheet = cuesheet_block
				self._blocks.append(cuesheet_block)
			elif block_type == FLACMetadataBlockType.PICTURE:
				picture = FLACPicture.load(metadata_block_data)
				self.pictures.append(picture)
				self._blocks.append(picture)
			elif block_type >= 127:
				raise InvalidBlock(f"{block_type} is not a valid FLAC metadata block type.")
			else:
				self._blocks.append(FLACMetadataBlock(block_type, metadata_block_data))

			if is_last_block:
				pos = self._obj.tell()
				self.streaminfo._start = pos
				self.streaminfo._size = self.filesize - self.streaminfo._start

				if self.streaminfo.duration:
					self.streaminfo.bitrate = self.streaminfo._size * 8 / self.streaminfo.duration
				else:
					self.streaminfo.bitrate = 0

				break
			else:
				header_data = self._obj.read(4)

		self._obj.close()

		return self



from .flac import *
from .id3v1 import *
from .id3v2 import *
from .id3v2_frames import *
from .mp3 import *
from .tables import *
from .vorbis import *
from .wav import *


__all__ = [
	*flac.__all__,
	*id3v1.__all__,
	*id3v2_frames.__all__,
	*id3v2.__all__,
	*mp3.__all__,
	*tables.__all__,
	*vorbis.__all__,
	*wav.__all__,
]



__all__ = [
	'determine_format',
	'load',
	'loads',
]

import os
from io import (
	BufferedReader,
	FileIO,
)

from .exceptions import (
	InvalidFormat,
	UnsupportedFormat,
)
from .formats import (
	FLAC,
	MP3,
	WAV,
	ID3v2,
	MP3StreamInfo,
)
from .utils import DataReader


def determine_format(data):
	"""Determine the format of an audio file.

	Parameters:
		data (bytes-like object, str, os.PathLike, or file-like object):
			A bytes-like object, filepath, path-like object
			or file-like object of an audio file.

	Returns:
		Format: An audio format class if supported, else None.
	"""

	# Only convert if not already a DataReader.
	# Otherwise ``find_mpeg_frames`` caching won't work.
	if not isinstance(data, DataReader):
		try:
			data = DataReader(data)
		except AttributeError:
			return None

	data.seek(0, os.SEEK_SET)
	d = data.peek(4)

	if d.startswith(b'fLaC'):
		return FLAC

	if d.startswith(b'RIFF'):
		return WAV

	if d.startswith(b'ID3'):
		ID3v2.load(data)

	if data.peek(4) == b'fLaC':
		return FLAC

	try:
		MP3StreamInfo.find_mpeg_frames(data)
	except InvalidFormat:
		return None
	else:
		return MP3


def load(f):
	"""Load audio metadata from filepath or file-like object.

	Parameters:
		f (str, os.PathLike, or file-like object):
			A filepath, path-like object or file-like object of an audio file.

	Returns:
		Format: An audio format object.

	Raises:
		UnsupportedFormat: If file is not of a supported format.
		ValueError: If filepath/file-like object is not valid or readable.
	"""

	if (
		not isinstance(f, (os.PathLike, str))
		and not (
			isinstance(f, BufferedReader)
			and isinstance(f.raw, FileIO)
		)
	):
		raise ValueError("Not a valid filepath or file-like object.")

	data = DataReader(f)

	parser_cls = determine_format(data)

	if parser_cls is None:
		raise UnsupportedFormat("Supported format signature not found.")
	else:
		data.seek(0, os.SEEK_SET)

	return parser_cls.load(data)


def loads(b):
	"""Load audio metadata from a bytes-like object.

	Parameters:
		b (bytes-like object): A bytes-like object of an audio file.

	Returns:
		Format: An audio format object.

	Raises:
		UnsupportedFormat: If file is not of a supported format.
	"""

	try:
		memoryview(b)
	except TypeError:
		raise ValueError("Not a valid bytes-like object.")

	data = DataReader(b)

	parser_cls = determine_format(data)

	if parser_cls is None:
		raise UnsupportedFormat("Supported format signature not found.")
	else:
		data.seek(0, os.SEEK_SET)

	return parser_cls.load(data)



__all__ = [
	'FLACMetadataBlockType',
	'ID3PictureType',
	'ID3Version',
	'ID3v1Genres',
	'ID3v2FrameIDs',
	'LAMEBitrateMode',
	'LAMEChannelMode',
	'LAMEPreset',
	'LAMEReplayGainOrigin',
	'LAMEReplayGainType',
	'LAMESurroundInfo',
	'MP3BitrateMode',
	'MP3Bitrates',
	'MP3ChannelMode',
	'MP3SampleRates',
	'MP3SamplesPerFrame',
]

from enum import (
	Enum,
	IntEnum,
)


class _BaseEnum(Enum):  # pragma: nocover
	def __repr__(self):
		return f'<{self.__class__.__name__}.{self.name}>'


class _BaseIntEnum(IntEnum):  # pragma: nocover
	def __repr__(self):
		return f'<{self.__class__.__name__}.{self.name}>'


# https://xiph.org/flac/format.html#metadata_block_header
class FLACMetadataBlockType(_BaseIntEnum):
	STREAMINFO = 0
	PADDING = 1
	APPLICATION = 2
	SEEKTABLE = 3
	VORBIS_COMMENT = 4
	CUESHEET = 5
	PICTURE = 6


# http://id3.org/id3v2.3.0#Attached_picture
class ID3PictureType(_BaseIntEnum):
	OTHER = 0
	FILE_ICON = 1
	OTHER_FILE_ICON = 2
	COVER_FRONT = 3
	COVER_BACK = 4
	LEAFLET_PAGE = 5
	MEDIA = 6
	LEAD_ARTIST = 7
	ARTIST = 8
	CONDUCTOR = 9
	BAND = 10
	COMPOSER = 11
	LYRICIST = 12
	RECORDING_LOCATION = 13
	DURING_RECORDING = 14
	DURING_PERFORMANCE = 15
	SCREEN_CAPTURE = 16
	FISH = 17
	ILLUSTRATION = 18
	ARTIST_LOGOTYPE = 19
	BAND_LOGOTYPE = 19
	PUBLISHER_LOGOTYPE = 20
	STUDIO_LOGOTYPE = 20


class ID3Version(_BaseEnum):
	v10 = (1, 0)
	v11 = (1, 1)
	v22 = (2, 2)
	v23 = (2, 3)
	v24 = (2, 4)


# https://en.wikipedia.org/wiki/List_of_ID3v1_Genres
ID3v1Genres = [
	'Blues',
	'Classic Rock',
	'Country',
	'Dance',
	'Disco',
	'Funk',
	'Grunge',
	'Hip-Hop',
	'Jazz',
	'Metal',
	'New Age',
	'Oldies',
	'Other',
	'Pop',
	'R&B',
	'Rap',
	'Reggae',
	'Rock',
	'Techno',
	'Industrial',
	'Alternative',
	'Ska',
	'Death Metal',
	'Pranks',
	'Soundtrack',
	'Euro-Techno',
	'Ambient',
	'Trip-Hop',
	'Vocal',
	'Jazz+Funk',
	'Fusion',
	'Trance',
	'Classical',
	'Instrumental',
	'Acid',
	'House',
	'Game',
	'Sound Clip',
	'Gospel',
	'Noise',
	'Alt Rock',
	'Bass',
	'Soul',
	'Punk',
	'Space',
	'Meditative',
	'Instrumental Pop',
	'Instrumental Rock',
	'Ethnic',
	'Gothic',
	'Darkwave',
	'Techno-Industrial',
	'Electronic',
	'Pop-Folk',
	'Eurodance',
	'Dream',
	'Southern Rock',
	'Comedy',
	'Cult',
	'Gangsta Rap',
	'Top 40',
	'Christian Rap',
	'Pop/Funk',
	'Jungle',
	'Native American',
	'Cabaret',
	'New Wave',
	'Psychedelic',
	'Rave',
	'Showtunes',
	'Trailer',
	'Lo-Fi',
	'Tribal',
	'Acid Punk',
	'Acid Jazz',
	'Polka',
	'Retro',
	'Musical',
	'Rock & Roll',
	'Hard Rock',
	'Folk',
	'Folk-Rock',
	'National Folk',
	'Swing',
	'Fast-Fusion',
	'Bebop',
	'Latin',
	'Revival',
	'Celtic',
	'Bluegrass',
	'Avantgarde',
	'Gothic Rock',
	'Progressive Rock',
	'Symphonic Rock',
	'Slow Rock',
	'Big Band',
	'Chorus',
	'Easy Listening',
	'Acoustic',
	'Humour',
	'Speech',
	'Chanson',
	'Opera',
	'Chamber Music',
	'Sonata',
	'Symphony',
	'Booty Bass',
	'Primus',
	'Porn Groove',
	'Satire',
	'Slow Jam',
	'Club',
	'Tango',
	'Samba',
	'Folklore',
	'Ballad',
	'Power Ballad',
	'Rhythmic Soul',
	'Freestyle',
	'Duet',
	'Punk Rock',
	'Drum Solo',
	'A Cappella',
	'Euro-House',
	'Dance Hall',
	'Goa',
	'Drum & Bass',
	'Club-House',
	'Hardcore',
	'Terror',
	'Indie',
	'BritPop',
	'Afro-Punk',
	'Polsk Punk',
	'Beat',
	'Christian Gangsta Rap',
	'Heavy Metal',
	'Black Metal',
	'Crossover',
	'Contemporary Christian',
	'Chrstian Rock',
	'Merengue',
	'Salsa',
	'Thrash Metal',
	'Anime',
	'JPop',
	'Synthpop',
	'Abstract',
	'Art Rock',
	'Baroque',
	'Bhangra',
	'Big Beat',
	'Breakbeat',
	'Chillout',
	'Downtempo',
	'Dub',
	'EBM',
	'Eclectic',
	'Electro',
	'Electroclash',
	'Emo',
	'Experimental',
	'Garage',
	'Global',
	'IDM',
	'Illibient',
	'Industro-Goth',
	'Jam Band',
	'Krautrock',
	'Leftfield',
	'Lounge',
	'Math Rock',
	'New Romantic',
	'Nu-Breakz',
	'Post-Punk',
	'Post-Rock',
	'Psytrance',
	'Shoegaze',
	'Space Rock',
	'Trop Rock',
	'World Music',
	'Neoclassical',
	'Audiobook',
	'Audio Theatre',
	'Neue Deutsche Welle',
	'Podcast',
	'Indie Rock',
	'G-Funk',
	'Dubstep',
	'Garage Rock',
	'Psybient',
]


ID3v2FrameIDs = {
	# http://id3.org/id3v2-00
	ID3Version.v22: {
		'BUF', 'CNT', 'COM', 'CRA', 'CRM', 'ETC', 'EQU', 'IPL',
		'LNK', 'MCI', 'MLL', 'PIC', 'POP', 'REV', 'RVA', 'SLT',
		'STC', 'TAL', 'TBP', 'TCM', 'TCO', 'TCR', 'TDA', 'TDY',
		'TEN', 'TFT', 'TIM', 'TKE', 'TLA', 'TLE', 'TMT', 'TOA',
		'TOF', 'TOL', 'TOR', 'TOT', 'TP1', 'TP2', 'TP3', 'TP4',
		'TPA', 'TPB', 'TRC', 'TRD', 'TRK', 'TSI', 'TSS', 'TT1',
		'TT2', 'TT3', 'TXT', 'TXX', 'TYE', 'UFI', 'ULT', 'WAF',
		'WAR', 'WAS', 'WCM', 'WCP', 'WPB', 'WXX',
	},
	# http://id3.org/id3v2.3.0#Declared_ID3v2_frames
	ID3Version.v23: {
		'AENC', 'APIC', 'COMM', 'COMR', 'ENCR', 'EQUA', 'ETCO',
		'GEOB', 'GRID', 'IPLS', 'LINK', 'MCDI', 'MLLT', 'OWNE',
		'PRIV', 'PCNT', 'POPM', 'POSS', 'RBUF', 'RVAD', 'RVRB',
		'SYLT', 'SYTC', 'TALB', 'TBPM', 'TCOM', 'TCON', 'TCOP',
		'TDAT', 'TDLY', 'TENC', 'TEXT', 'TFLT', 'TIME', 'TIT1',
		'TIT2', 'TIT3', 'TKEY', 'TLAN', 'TLEN', 'TMED', 'TOAL',
		'TOFN', 'TOLY', 'TOPE', 'TORY', 'TOWN', 'TPE1', 'TPE2',
		'TPE3', 'TPE4', 'TPOS', 'TPUB', 'TRCK', 'TRDA', 'TRSN',
		'TRSO', 'TSIZ', 'TSRC', 'TSSE', 'TYER', 'TXXX', 'UFID',
		'USER', 'USLT', 'WCOM', 'WCOP', 'WOAF', 'WOAR', 'WOAS',
		'WORS', 'WPAY', 'WPUB', 'WXXX',
	},
	# http://id3.org/id3v2.4.0-frames
	ID3Version.v24: {
		'AENC', 'APIC', 'ASPI', 'COMM', 'COMR', 'ENCR', 'EQU2',
		'ETCO', 'GEOB', 'GRID', 'LINK', 'MCDI', 'MLLT', 'OWNE',
		'PRIV', 'PCNT', 'POPM', 'POSS', 'RBUF', 'RVA2', 'RVRB',
		'SEEK', 'SIGN', 'SYLT', 'SYTC', 'TALB', 'TBPM', 'TCOM',
		'TCON', 'TCOP', 'TDEN', 'TDLY', 'TDOR', 'TDRC', 'TDRL',
		'TDTG', 'TENC', 'TEXT', 'TFLT', 'TIPL', 'TIT1', 'TIT2',
		'TIT3', 'TKEY', 'TLAN', 'TLEN', 'TMCL', 'TMED', 'TMOO',
		'TOAL', 'TOFN', 'TOLY', 'TOPE', 'TOWN', 'TPE1', 'TPE2',
		'TPE3', 'TPE4', 'TPOS', 'TPRO', 'TPUB', 'TRCK', 'TRSN',
		'TRSO', 'TSOA', 'TSOP', 'TSOT', 'TSRC', 'TSSE', 'TSST',
		'TXXX', 'UFID', 'USER', 'USLT', 'WCOM', 'WCOP', 'WOAF',
		'WOAR', 'WOAS', 'WORS', 'WPAY', 'WPUB', 'WXXX',
	}
}


class LAMEBitrateMode(_BaseIntEnum):
	UNKNOWN = 0
	CBR = 1
	ABR = 2
	VBR_METHOD_1 = 3
	VBR_METHOD_2 = 4
	VBR_METHOD_3 = 5
	VBR_METHOD_4 = 6
	CBR_2_PASS = 8
	ABR_2_PASS = 9
	RESERVED = 15


class LAMEChannelMode(_BaseIntEnum):
	MONO = 0
	STEREO = 1
	DUAL_CHANNEL = 2
	JOINT_STEREO = 3
	FORCED = 4
	AUTO = 5
	INTENSITY = 6
	UNDEFINED = 7


LAMEPreset = _BaseIntEnum(
	'LAMEPreset',
	[
		('Unknown', 0),
		*[(f'ABR{x}', x) for x in range(8, 321)],
		('V9', 410),
		('V8', 420),
		('V7', 430),
		('V6', 440),
		('V5', 450),
		('V4', 460),
		('V3', 470),
		('V2', 480),
		('V1', 490),
		('V0', 500),
		('r3mix', 1000),
		('standard', 1001),
		('extreme', 1002),
		('insane', 1003),
		('standard_fast', 1004),
		('extreme_fast', 1005),
		('medium', 1006),
		('medium_fast', 1007),
	],
)


class LAMEReplayGainOrigin(_BaseIntEnum):
	NOT_SET = 0
	ARTIST = 1
	USER = 2
	MODEL = 3
	AVERAGE = 4


class LAMEReplayGainType(_BaseIntEnum):
	NOT_SET = 0
	RADIO = 1
	AUDIOPHILE = 2


class LAMESurroundInfo(_BaseIntEnum):
	NO_SURROUND = 0
	DPL = 1
	DPL2 = 2
	AMBISONIC = 3


# (version, layer): bitrate in kilobits per second
MP3Bitrates = {
	(1, 1): [0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448],
	(1, 2): [0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384],
	(1, 3): [0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320],
	(2, 1): [0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256],
	(2, 2): [0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
	(2, 3): [0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
	(2.5, 1): [0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256],
	(2.5, 2): [0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
	(2.5, 3): [0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
}


class MP3BitrateMode(_BaseIntEnum):
	UNKNOWN = 0
	CBR = 1
	VBR = 2
	ABR = 3


class MP3ChannelMode(_BaseIntEnum):
	STEREO = 0
	JOINT_STEREO = 1
	DUAL_CHANNEL = 2
	MONO = 3


# version
MP3SampleRates = {
	1: [44100, 48000, 32000],
	2: [22050, 24000, 16000],
	2.5: [11025, 12000, 8000],
}


# (version, layer): (samples_per_frame, slot_size)
MP3SamplesPerFrame = {
	(1, 1): (384, 4),
	(1, 2): (1152, 1),
	(1, 3): (1152, 1),
	(2, 1): (384, 4),
	(2, 2): (1152, 1),
	(2, 3): (576, 1),
	(2.5, 1): (384, 4),
	(2.5, 2): (1152, 1),
	(2.5, 3): (576, 1),
}



# http://id3.org/Developer%20Information

__all__ = [
	'ID3v2',
	'ID3v2Flags',
	'ID3v2Frames',
	'ID3v2Header',
]

import struct
import warnings
from collections import defaultdict

from attr import (
	attrib,
	attrs,
)
from bidict import frozenbidict
from tbm_utils import AttrMapping

from .id3v2_frames import *
from .tables import (
	ID3Version,
	ID3v2FrameIDs,
)
from ..exceptions import (
	InvalidFrame,
	InvalidHeader,
)
from ..models import Tags
from ..utils import (
	datareader,
	decode_synchsafe_int,
)

try:  # pragma: nocover
	import bitstruct.c as bitstruct
except ImportError:  # pragma: nocover
	import bitstruct


# Mappings used: https://picard.musicbrainz.org/docs/mappings/
class ID3v2Frames(Tags):
	_v22_FIELD_MAP = frozenbidict(
		{
			'album': 'TAL',
			'albumartist': 'TP2',
			'artist': 'TP1',
			'audiodelay': 'TDY',
			'audiolength': 'TLE',
			'audiosize': 'TSI',
			'bpm': 'TBP',
			'comment': 'COM',
			'composer': 'TCM',
			'conductor': 'TP3',
			'copyright': 'TCR',
			'date': 'TYE',
			'discnumber': 'TPA',
			'encodedby': 'TEN',
			'encodersettings': 'TSS',
			'genre': 'TCO',
			'grouping': 'TT1',
			'isrc': 'TRC',
			'label': 'TPB',
			'language': 'TLA',
			'lyricist': 'TXT',
			'lyrics': 'ULT',
			'media': 'TMT',
			'originalalbum': 'TOT',
			'originalartist': 'TOA',
			'originalauthor': 'TOL',
			'originaldate': 'TOR',
			'pictures': 'PIC',
			'playcount': 'CNT',
			'remixer': 'TP4',
			'subtitle': 'TT3',
			'title': 'TT2',
			'tracknumber': 'TRK',
		},
	)

	_v23_FIELD_MAP = frozenbidict(
		{
			'album': 'TALB',
			'albumsort': 'TSOA',
			'albumartist': 'TPE2',
			'albumartistsort': 'TSO2',
			'artist': 'TPE1',
			'artistsort': 'TSOP',
			'audiodelay': 'TDLY',
			'audiolength': 'TLEN',
			'audiosize': 'TSIZ',
			'bpm': 'TBPM',
			'comment': 'COMM',
			'compilation': 'TCMP',
			'composer': 'TCOM',
			'composersort': 'TSOC',
			'conductor': 'TPE3',
			'copyright': 'TCOP',
			'date': 'TYER',
			'discnumber': 'TPOS',
			'encodedby': 'TENC',
			'encodersettings': 'TSSE',
			'genre': 'TCON',
			'grouping': 'TIT1',
			'isrc': 'TSRC',
			'label': 'TPUB',
			'language': 'TLAN',
			'lyricist': 'TEXT',
			'lyrics': 'USLT',
			'media': 'TMED',
			'originalalbum': 'TOAL',
			'originalartist': 'TOPE',
			'originalauthor': 'TOLY',
			'originaldate': 'TORY',
			'pictures': 'APIC',
			'playcount': 'PCNT',
			'remixer': 'TPE4',
			'subtitle': 'TIT3',
			'title': 'TIT2',
			'titlesort': 'TSOT',
			'tracknumber': 'TRCK',
		},
	)

	_v24_FIELD_MAP = frozenbidict(
		{
			'album': 'TALB',
			'albumsort': 'TSOA',
			'albumartist': 'TPE2',
			'albumartistsort': 'TSO2',
			'artist': 'TPE1',
			'artistsort': 'TSOP',
			'audiodelay': 'TDLY',
			'audiolength': 'TLEN',
			'audiosize': 'TSIZ',
			'bpm': 'TBPM',
			'comment': 'COMM',
			'compilation': 'TCMP',
			'composer': 'TCOM',
			'composersort': 'TSOC',
			'conductor': 'TPE3',
			'copyright': 'TCOP',
			'date': 'TDRC',
			'discnumber': 'TPOS',
			'encodedby': 'TENC',
			'encodersettings': 'TSSE',
			'genre': 'TCON',
			'grouping': 'TIT1',
			'isrc': 'TSRC',
			'label': 'TPUB',
			'language': 'TLAN',
			'lyricist': 'TEXT',
			'lyrics': 'USLT',
			'media': 'TMED',
			'mood': 'TMOO',
			'originalalbum': 'TOAL',
			'originalartist': 'TOPE',
			'originalauthor': 'TOLY',
			'originaldate': 'TDOR',
			'pictures': 'APIC',
			'playcount': 'PCNT',
			'remixer': 'TPE4',
			'subtitle': 'TIT3',
			'title': 'TIT2',
			'titlesort': 'TSOT',
			'tracknumber': 'TRCK',
		},
	)

	def __init__(self, mapping=None, *, id3_version=ID3Version.v24, **kwargs):
		self._version = ID3Version(id3_version)

		if self._version is ID3Version.v22:
			self.FIELD_MAP = ID3v2Frames._v22_FIELD_MAP
		elif self._version is ID3Version.v23:
			self.FIELD_MAP = ID3v2Frames._v23_FIELD_MAP
		elif self._version is ID3Version.v24:
			self.FIELD_MAP = ID3v2Frames._v24_FIELD_MAP
		else:
			raise ValueError(f"Unsupported ID3 version: {id3_version}")

		super().__init__(mapping, **kwargs)

	@datareader
	@classmethod
	def load(cls, data, id3_version):
		id3_version = ID3Version(id3_version)
		if id3_version is ID3Version.v22:
			struct_pattern = '3s3B'
			size_len = 3
			per_byte = 8
		elif id3_version is ID3Version.v23:
			struct_pattern = '4s4B2B'
			size_len = 4
			per_byte = 8
		elif id3_version is ID3Version.v24:
			struct_pattern = '4s4B2B'
			size_len = 4
			per_byte = 7
		else:
			raise ValueError(f"Unsupported ID3 version: {id3_version}")  # pragma: nocover

		frames = defaultdict(list)
		while True:
			try:
				frame = ID3v2Frame.load(data, struct_pattern, size_len, per_byte)
			except InvalidFrame:
				break

			# Ignore oddities/bad frames.
			if frame is None:
				continue

			# TODO: Handle commonly seen frames like from iTunes.
			# Ignore frames not defined in spec for ID3 version.
			# Warn user and encourage reporting.
			if frame.id not in ID3v2FrameIDs[id3_version]:
				warnings.warn(
					f"Ignoring '{frame.id}' frame with value '{frame.value}'.\n"
					f"'{frame.id}' is not supported in the ID3v2.{id3_version.value[1]} specification.\n"
				)
				continue

			# TODO: Finish any missing frame types.
			# TODO: Move representation into frame classes?
			if isinstance(
				frame,
				(
					ID3v2CommentFrame,
					ID3v2SynchronizedLyricsFrame,
					ID3v2UnsynchronizedLyricsFrame,
				),
			):
				frames[f'{frame.id}:{frame.description}:{frame.language}'].append(frame.value)
			elif isinstance(frame, ID3v2GenreFrame):
				frames['TCON'] = frame.value
			elif isinstance(frame, ID3v2GEOBFrame):
				frames[f'GEOB:{frame.description}'].append(
					{

						'filename': frame.filename,
						'mime_type': frame.mime_type,
						'value': frame.value,
					},
				)
			elif isinstance(frame, ID3v2PrivateFrame):
				frames[f'PRIV:{frame.owner}'].append(frame.value)
			elif isinstance(
				frame,
				(
					ID3v2UserTextFrame,
					ID3v2UserURLLinkFrame,
				),
			):
				frames[f'{frame.id}:{frame.description}'].append(frame.value)
			elif isinstance(
				frame,
				(
					ID3v2NumericTextFrame,
					ID3v2TextFrame,
					ID3v2TimestampFrame,
				),
			):
				frames[frame.id] = frame.value
			else:
				frames[frame.id].append(frame.value)

		return cls(frames, id3_version=id3_version)


@attrs(repr=False)
class ID3v2Flags(AttrMapping):
	unsync = attrib(converter=bool)
	extended = attrib(converter=bool)
	experimental = attrib(converter=bool)
	footer = attrib(converter=bool)


@attrs(repr=False)
class ID3v2Header(AttrMapping):
	_size = attrib()
	version = attrib()
	flags = attrib(converter=ID3v2Flags.from_mapping)

	def __repr__(self):
		repr_dict = {}

		for k, v in sorted(self.items()):
			if not k.startswith('_'):
				repr_dict[k] = v

		return super().__repr__(repr_dict=repr_dict)

	@datareader
	@classmethod
	def load(cls, data):
		if data.read(3) != b"ID3":
			raise InvalidHeader("Valid ID3v2 header not found.")

		major, revision, flags_, sync_size = struct.unpack('BBs4s', data.read(7))

		try:
			version = ID3Version((2, major))
		except ValueError:  # pragma: nocover
			raise ValueError(f"Unsupported ID3 version (2.{major}).")

		flags = bitstruct.unpack_dict(
			'b1 b1 b1 b1',
			[
				'unsync',
				'extended',
				'experimental',
				'footer',
			],
			flags_,
		)

		size = decode_synchsafe_int(sync_size, 7)

		return cls(size, version, flags)


class ID3v2(AttrMapping):
	def __repr__(self):
		repr_dict = {}

		for k, v in sorted(self.items()):
			if not k.startswith('_'):
				repr_dict[k] = v

		return super().__repr__(repr_dict=repr_dict)

	@datareader
	@classmethod
	def load(cls, data):
		if data.peek(3) != b"ID3":
			raise InvalidHeader("Valid ID3v2 header not found.")

		self = cls()

		self._header = ID3v2Header.load(data.read(10))
		self._size = 10 + self._header._size

		if self._header.flags.extended:
			ext_size = decode_synchsafe_int(
				struct.unpack('4B', data.read(4))[0:4],
				7,
			)
			self._size += ext_size

			if self._header is ID3Version.v24:
				data.read(ext_size - 4)
			else:
				data.read(ext_size)

		if self._header.flags.footer:
			self._size += 10
			data.read(10)

		self.tags = ID3v2Frames.load(
			data.read(self._header._size),
			self._header.version,
		)
		self.pictures = self.tags.pop('pictures', [])

		return self



# http://soundfile.sapp.org/doc/WaveFormat/
# http://www-mmsp.ece.mcgill.ca/Documents/AudioFormats/WAVE/WAVE.html

__all__ = [
	'RIFFTags',
	'WAV',
	'WAVStreamInfo',
]

import os
import struct

from attr import (
	attrib,
	attrs,
)
from bidict import frozenbidict

from .id3v2 import ID3v2
from ..exceptions import (
	InvalidChunk,
	InvalidFrame,
	InvalidHeader,
)
from ..models import (
	Format,
	StreamInfo,
	Tags,
)
from ..utils import datareader


# https://www.recordingblogs.com/wiki/list-chunk-of-a-wave-file
class RIFFTags(Tags):
	FIELD_MAP = frozenbidict(
		{
			'album': 'IPRD',
			'artist': 'IART',
			'comment': 'ICMT',
			'copyright': 'ICOP',
			'date': 'ICRD',
			'encodedby': 'IENC',
			'genre': 'IGNR',
			'language': 'ILNG',
			'rating': 'IRTD',
			'title': 'INAM',
			'tracknumber': 'ITRK',
		},
	)

	@datareader
	@classmethod
	def load(cls, data):
		if data.read(4) != b'INFO':
			raise InvalidChunk('Valid RIFF INFO chunk not found.')

		fields = {}

		field = data.read(4)
		while len(field):
			size = struct.unpack('I', data.read(4))[0]
			value = data.read(size).strip(b'\x00').decode('utf-8')
			fields[field.decode('utf-8')] = [value]

			b = data.read(1)
			while b == b'\x00':
				b = data.read(1)

			if b:
				data.seek(-1, os.SEEK_CUR)

			field = data.read(4)

		return cls(fields)


@attrs(repr=False)
class WAVStreamInfo(StreamInfo):
	_start = attrib()
	_size = attrib()
	bit_depth = attrib()
	bitrate = attrib()
	channels = attrib()
	duration = attrib()
	sample_rate = attrib()


class WAV(Format):
	"""WAV file format object.

	Extends :class:`Format`.

	Attributes:
		pictures (list): A list of :class:`ID3v2Picture` objects.
		streaminfo (WAVStreamInfo): The audio stream information.
		tags (ID3v2Frames or RIFFTags): The ID3v2 or RIFF tags, if present.
	"""

	tags_type = RIFFTags

	@classmethod
	def load(cls, data):
		self = super()._load(data)

		chunk_id = self._obj.read(4)

		# chunk_size
		self._obj.read(4)

		format_ = self._obj.read(4)

		if chunk_id != b'RIFF' or format_ != b'WAVE':
			raise InvalidHeader("Valid WAVE header not found.")

		subchunk_header = self._obj.read(8)
		while len(subchunk_header) == 8:
			subchunk_id, subchunk_size = struct.unpack(
				'4sI',
				subchunk_header,
			)

			if subchunk_id == b'fmt ':
				audio_format, channels, sample_rate = struct.unpack(
					'HHI',
					self._obj.read(8),
				)

				byte_rate, block_align, bit_depth = struct.unpack(
					'<IHH',
					self._obj.read(8),
				)

				bitrate = byte_rate * 8

				self._obj.read(subchunk_size - 16)  # Read through rest of subchunk if not PCM.
			elif subchunk_id == b'data':
				audio_start = self._obj.tell()
				audio_size = subchunk_size
				self._obj.seek(subchunk_size, os.SEEK_CUR)
			elif (
				subchunk_id == b'LIST'
				and self._obj.peek(4) == b'INFO'
			):
				self._riff = RIFFTags.load(self._obj.read(subchunk_size))
			elif subchunk_id.lower() == b'id3 ':
				try:
					id3 = ID3v2.load(self._obj)
				except (InvalidFrame, InvalidHeader):
					raise
				else:
					self._id3 = id3
			else:
				# TODO
				self._obj.seek(subchunk_size, os.SEEK_CUR)  # pragma: nocover

			subchunk_header = self._obj.read(8)

		if '_id3' in self:
			self.pictures = self._id3.pictures
			self.tags = self._id3.tags
		elif '_riff' in self:
			self.tags = self._riff

		try:
			duration = audio_size / byte_rate

			self.streaminfo = WAVStreamInfo(
				audio_start,
				audio_size,
				bit_depth,
				bitrate,
				channels,
				duration,
				sample_rate,
			)
		except UnboundLocalError:
			raise InvalidHeader("Valid WAVE stream info not found.")

		self._obj.close()

		return self



__all__ = [
	'DataReader',
	'decode_synchsafe_int',
	'get_image_size',
	'humanize_bitrate',
	'humanize_sample_rate',
]

import os
import struct
from codecs import (
	BOM_UTF16_BE,
	BOM_UTF16_LE,
)
from functools import reduce
from io import (
	DEFAULT_BUFFER_SIZE,
	BufferedReader,
	BytesIO,
	FileIO,
)

import wrapt


class DataReader(BufferedReader):
	def __init__(self, data, buffer_size=DEFAULT_BUFFER_SIZE):
		if isinstance(data, (BufferedReader, DataReader)):
			if isinstance(data.raw, FileIO):
				data = FileIO(data.name, 'rb')
			else:
				data = BytesIO(data.read())
		elif isinstance(data, (os.PathLike, str)):
			data = FileIO(data, 'rb')
		elif isinstance(data, (bytearray, bytes, memoryview)):
			data = BytesIO(data)

		super().__init__(data, buffer_size=buffer_size)

	def peek(self, size=DEFAULT_BUFFER_SIZE):
		if size > DEFAULT_BUFFER_SIZE:
			size = DEFAULT_BUFFER_SIZE

		peeked = super().peek(size)[:size]

		if len(peeked) < size:
			peeked = self.read(size)
			self.seek(-len(peeked), os.SEEK_CUR)

		return peeked


@wrapt.decorator
def datareader(wrapped, instance, args, kwargs):
	if not isinstance(args[0], DataReader):
		data = DataReader(args[0])
		args = (data, *args[1:])

	return wrapped(*args, **kwargs)


def decode_bytestring(b, encoding='iso-8859-1'):
	if not b:
		return ''

	if encoding.startswith('utf-16'):
		if len(b) % 2 != 0 and b[-1:] == b'\x00':
			b = b[:-1]

		if b.startswith((BOM_UTF16_BE, BOM_UTF16_LE)):  # pragma: nobranch
			b = b[2:]

	return b.decode(encoding).rstrip('\x00')


def decode_synchsafe_int(data, per_byte):
	return reduce(lambda value, element: (value << per_byte) + element, data, 0)


def determine_encoding(b):
	first = b[0:1]

	if first == b'\x00':
		encoding = 'iso-8859-1'
	elif first == b'\x01':
		encoding = 'utf-16-be' if b[1:3] == BOM_UTF16_BE else 'utf-16-le'
	elif first == b'\x02':
		encoding = 'utf-16-be'
	elif first == b'\x03':
		encoding = 'utf-8'
	else:
		encoding = 'iso-8859-1'

	return encoding


def get_image_size(data):
	if not isinstance(data, DataReader):  # pragma: nocover
		data = DataReader(data)

	b = data.read(56)
	size = len(b)

	width = height = 0
	if size >= 10 and b[:6] in [b'GIF87a', b'GIF89a']:
		width, height = struct.unpack('<hh', b[6:10])
	elif size >= 24 and b.startswith(b'\x89PNG') and b[12:16] == b'IHDR':
		width, height = struct.unpack('>LL', b[16:24])
	elif size >= 2 and b.startswith(b'\xff\xd8'):
		data.seek(0)

		size = 2
		ftype = 0
		while not 0xc0 <= ftype <= 0xcf or ftype in [0xc4, 0xc8, 0xcc]:
			data.seek(size, os.SEEK_CUR)
			while True:
				b = data.read(1)
				if b != b'\xff':
					break

			ftype = ord(b)
			size = struct.unpack('>H', data.read(2))[0] - 2

		data.seek(1, os.SEEK_CUR)
		height, width = struct.unpack('>HH', data.read(4))
	elif size >= 12 and b.startswith(b'\x00\x00\x00\x0cjP'):
		height, width = struct.unpack('>LL', b[48:])
	else:
		raise ValueError(f"'data' is not a supported image file.")

	return width, height


def humanize_bitrate(bitrate):
	for divisor, symbol in [(1000 ** 1, 'Kbps'), (1, 'bps')]:
		if bitrate >= divisor:
			break

	return f'{round(bitrate / divisor)} {symbol}'


def humanize_sample_rate(sample_rate):
	for divisor, symbol in [(1000 ** 1, 'KHz'), (1, 'Hz')]:
		if sample_rate >= divisor:
			break

	value = sample_rate / divisor

	return f'{value if value.is_integer() else value:.1f} {symbol}'


def split_encoded(data, encoding):
	try:
		if encoding in ['iso-8859-1', 'utf-8']:
			head, tail = data.split(b'\x00', 1)
		else:
			if len(data) % 2 != 0:
				data += b'\x00'

			head, tail = data.split(b'\x00\x00', 1)

			if len(head) % 2 != 0:
				head, tail = data.split(b'\x00\x00\x00', 1)
				head += b'\x00'
	except ValueError:
		return (data,)

	return head, tail



__all__ = [
	'Format',
	'Picture',
	'StreamInfo',
	'Tags',
]

import os

from bidict import frozenbidict
from tbm_utils import (
	AttrMapping,
	humanize_duration,
	humanize_filesize,
)

from .utils import (
	datareader,
	humanize_bitrate,
	humanize_sample_rate,
)


class Tags(AttrMapping):
	FIELD_MAP = frozenbidict()

	def __getitem__(self, key):
		k = self.FIELD_MAP.get(key, key)

		return super().__getitem__(k)

	def __setitem__(self, key, value):
		k = self.FIELD_MAP.get(key, key)

		super().__setitem__(k, value)

	def __delitem__(self, key):
		k = self.FIELD_MAP.get(key, key)

		super().__delitem__(k)

	def __iter__(self):
		return iter(
			self.FIELD_MAP.inv.get(k, k)
			for k in self.__dict__
			if not k.startswith('_') and not k == 'FIELD_MAP'
		)

	def __repr__(self, repr_dict=None):
		repr_dict = {
			self.FIELD_MAP.inv.get(k, k): v
			for k, v in self.__dict__.items()
			if not k.startswith('_') and not k == 'FIELD_MAP'
		}

		return super().__repr__(repr_dict=repr_dict)


class Format(AttrMapping):
	"""Base class for audio format objects.

	Attributes:
		filepath (str): Path to audio file, if applicable.
		filesize (int): Size of audio file.
		pictures (list): A list of :class:`Picture` objects.
		tags (Tags): A :class:`Tags` object.
	"""

	tags_type = Tags

	def __init__(self):
		self.filepath = None
		self.filesize = None
		self.pictures = []
		self.tags = self.tags_type()

	def __repr__(self):
		repr_dict = {}

		for k, v in sorted(self.items()):
			if (
				k == 'filesize'
				and v is not None
			):
				repr_dict[k] = humanize_filesize(v, precision=2)
			elif not k.startswith('_'):
				repr_dict[k] = v

		return super().__repr__(repr_dict=repr_dict)

	@datareader
	@classmethod
	def _load(cls, data):
		self = cls()
		self._obj = data

		try:
			self.filepath = os.path.abspath(self._obj.name)
			self.filesize = os.path.getsize(self._obj.name)
		except AttributeError:
			self.filepath = None
			self.filesize = len(self._obj.raw.getbuffer())

		return self


class Picture(AttrMapping):
	def __repr__(self):
		repr_dict = {}

		for k, v in sorted(self.items()):
			if k == 'data':
				repr_dict[k] = humanize_filesize(len(v), precision=2)
			elif not k.startswith('_'):
				repr_dict[k] = v

		return super().__repr__(repr_dict=repr_dict)


class StreamInfo(AttrMapping):
	def __repr__(self):
		repr_dict = {}

		for k, v in sorted(self.items()):
			if k == 'bitrate':
				repr_dict[k] = humanize_bitrate(v)
			elif k == 'duration':
				repr_dict[k] = humanize_duration(v)
			elif k == 'sample_rate':
				repr_dict[k] = humanize_sample_rate(v)
			elif not k.startswith('_'):
				repr_dict[k] = v

		return super().__repr__(repr_dict=repr_dict)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['audio_metadata', 'audio_metadata.formats']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=18.2,<19.4',
 'bidict<1.0.0',
 'bitstruct>=6.0,<9.0',
 'more-itertools>=4.0,<9.0',
 'pprintpp<1.0.0',
 'tbm-utils>=1.1.1,<2.0.0',
 'wrapt>=1.0,<2.0']

extras_require = \
{'dev': ['coverage[toml]>=5.0,<6.0',
         'flake8>=3.5,<4.0',
         'flake8-builtins>=1.0,<2.0',
         'flake8-comprehensions>=2.0,<=4.0',
         'flake8-import-order>=0.18,<0.19',
         'flake8-import-order-tbm>=1.0,<2.0',
         'nox>=2019,<2020',
         'pytest>=4.0,<6.0',
         'sphinx>=2.0,<3.0',
         'sphinx-material<1.0.0'],
 'doc': ['sphinx>=2.0,<3.0', 'sphinx-material<1.0.0'],
 'lint': ['flake8>=3.5,<4.0',
          'flake8-builtins>=1.0,<2.0',
          'flake8-comprehensions>=2.0,<=4.0',
          'flake8-import-order>=0.18,<0.19',
          'flake8-import-order-tbm>=1.0,<2.0'],
 'test': ['coverage[toml]>=5.0,<6.0', 'nox>=2019,<2020', 'pytest>=4.0,<6.0']}

setup_kwargs = {
    'name': 'audio-metadata',
    'version': '0.7.0',
    'description': 'A library for reading and, in the future, writing metadata from audio files.',
    'long_description': "# audio-metadata\n\n[![PyPI](https://img.shields.io/pypi/v/audio-metadata.svg?label=PyPI)](https://pypi.org/project/audio-metadata/)\n![](https://img.shields.io/badge/Python-3.6%2B-blue.svg)  \n[![GitHub CI](https://img.shields.io/github/workflow/status/thebigmunch/audio-metadata/CI?label=GitHub%20CI)](https://github.com/thebigmunch/audio-metadata/actions?query=workflow%3ACI)\n[![Codecov](https://img.shields.io/codecov/c/github/thebigmunch/audio-metadata.svg?label=Codecov)](https://codecov.io/gh/thebigmunch/audio-metadata)  \n[![Docs - Stable](https://img.shields.io/readthedocs/audio-metadata/stable.svg?label=Docs%20%28Stable%29)](https://audio-metadata.readthedocs.io/en/stable/)\n[![Docs - Latest](https://img.shields.io/readthedocs/audio-metadata/latest.svg?label=Docs%20%28Latest%29)](https://audio-metadata.readthedocs.io/en/latest/)\n\n[audio-metadata](https://github.com/thebigmunch/audio-metadata) is a library for reading and, in the future, writing audio metadata.\n\n### Why another audio metadata library? / Why not just use mutagen et al?\n\nClean and understandable code, nice API, and good UX (user experience) are the focal points of audio-metadata.\nOne or more of these things I feel are lacking from already existing alternatives\nenough to want to write something from scratch with them in mind.\nAlso, there are certain features not present in other solutions that would be prohibitively painful to add.\n\n\n### So, why should I use it?\n\nFeatures and functionality that set it apart:\n\n* Uses the Python standard load(s)/dump(s) API.\n\t* Can load filepaths, file-like objects and binary data (bytes-like objects).\n* Metadata objects look like a dict **and** act like a dict.\n\t* Some common libraries shadow the representation of a dict\n\t  and/or dict methods but do not behave like a dict.\n\t* Supports attribute-style access that can be mixed with dict key-subscription.\n* All metadata objects have user-friendly representations.\n\t* This includes *humanized* representations of certain values\n\t  like filesize, bitrate, duration, and sample rate.\n\n```\n>>> import audio_metadata\n\n>>> metadata = audio_metadata.load('05 - Heart of Hearts.flac')\n\n>>> metadata\n<FLAC ({\n\t'filepath': '05 - Heart of Hearts.flac',\n\t'filesize': '44.23 MiB',\n\t'pictures': [],\n\t'seektable': <FLACSeekTable (37 seekpoints)>,\n\t'streaminfo': <FLACStreamInfo ({\n\t\t'bit_depth': 16,\n\t\t'bitrate': '1022 Kbps',\n\t\t'channels': 2,\n\t\t'duration': '06:03',\n\t\t'md5': '3ae700893d099a5d281a5d8db7847671',\n\t\t'sample_rate': '44.1 KHz',\n\t})>,\n\t'tags': <VorbisComment ({\n\t\t'album': ['Myth Takes'],\n\t\t'artist': ['!!!'],\n\t\t'bpm': ['119'],\n\t\t'date': ['2007'],\n\t\t'genre': ['Dance Punk'],\n\t\t'title': ['Heart of Hearts'],\n\t\t'tracknumber': ['05'],\n\t})>,\n})>\n\n>>> metadata['streaminfo']\n<FLACStreamInfo ({\n\t'bit_depth': 16,\n\t'bitrate': '1022 Kbps',\n\t'channels': 2,\n\t'duration': '06:03',\n\t'md5': '3ae700893d099a5d281a5d8db7847671',\n\t'sample_rate': '44.1 KHz',\n})>\n\n>>> metadata.streaminfo.bitrate\n1022134.0362995076\n\n>>> metadata.streaminfo['duration']\n362.9066666666667\n\n>>> metadata['streaminfo'].sample_rate\n44100\n```\n\n\n## Installation\n\n``pip install -U audio-metadata``\n\n\n## Usage\n\nFor the release version, see the [stable docs](https://audio-metadata.readthedocs.io/en/stable/).  \nFor the development version, see the [latest docs](https://audio-metadata.readthedocs.io/en/latest/).\n\nThe high-level API and basic usage are covered, but more advanced features/functionality need documentation.\n\n\n## TODO\n\nPlease don't submit issues regarding the following:\n\n* Docs, docs, and more docs.\n* Finish ID3 abstractions.\n* Add support for more file formats.\n\t* MP4 reading is already well underway.\n* Add write support.\n\nIf you're willing to [contribute](https://github.com/thebigmunch/audio-metadata/blob/master/.github/CONTRIBUTING.md)\nyour time to work on them, post in the **Development** category on the [Discourse forum](https://forum.thebigmunch.me/).\n\n\n## Appreciation\n\nShowing appreciation is always welcome.\n\n#### Thank\n\n[![Say Thanks](https://img.shields.io/badge/thank-thebigmunch-blue.svg?style=flat-square)](https://saythanks.io/to/thebigmunch)\n\nGet your own thanks inbox at [SayThanks.io](https://saythanks.io/).\n\n#### Contribute\n\n[Contribute](https://github.com/thebigmunch/audio-metadata/blob/master/.github/CONTRIBUTING.md) by submitting bug reports, feature requests, or code.\n\n#### Help Others/Stay Informed\n\n[Discourse forum](https://forum.thebigmunch.me/)\n\n#### Referrals/Donations\n\n[![Digital Ocean](https://img.shields.io/badge/Digital_Ocean-referral-orange.svg?style=flat-square)](https://bit.ly/DigitalOcean-tbm-referral) [![Namecheap](https://img.shields.io/badge/Namecheap-referral-orange.svg?style=flat-square)](http://bit.ly/Namecheap-tbm-referral) [![PayPal](https://img.shields.io/badge/PayPal-donate-brightgreen.svg?style=flat-square)](https://bit.ly/PayPal-thebigmunch)\n",
    'author': 'thebigmunch',
    'author_email': 'mail@thebigmunch.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/thebigmunch/audio-metadata',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# http://id3.org/Developer%20Information

__all__ = [
	'ID3v2BaseFrame',
	'ID3v2CommentFrame',
	'ID3v2Frame',
	'ID3v2GEOBFrame',
	'ID3v2GenreFrame',
	'ID3v2NumberFrame',
	'ID3v2NumericTextFrame',
	'ID3v2Picture',
	'ID3v2PictureFrame',
	'ID3v2PrivateFrame',
	'ID3v2SynchronizedLyricsFrame',
	'ID3v2TextFrame',
	'ID3v2TimestampFrame',
	'ID3v2UnsynchronizedLyricsFrame',
	'ID3v2URLLinkFrame',
	'ID3v2UserTextFrame',
	'ID3v2UserURLLinkFrame',
	'ID3v2YearFrame',
]

import re
import string
import struct
from urllib.parse import unquote

from attr import (
	attrib,
	attrs,
)
from tbm_utils import AttrMapping

from .tables import (
	ID3PictureType,
	ID3v1Genres,
)
from ..exceptions import InvalidFrame
from ..models import Picture
from ..utils import (
	datareader,
	decode_bytestring,
	decode_synchsafe_int,
	determine_encoding,
	get_image_size,
	split_encoded,
)

_genre_re = re.compile(r"((?:\((?P<id>\d+|RX|CR)\))*)(?P<name>.+)?")


class ID3v2Picture(Picture):
	@datareader
	@classmethod
	def load(cls, data):
		data = data.read()

		encoding = determine_encoding(data[0:1])
		mime_start = 1
		mime_end = data.index(b'\x00', 1)
		mime_type = decode_bytestring(data[mime_start:mime_end])

		type_ = ID3PictureType(data[mime_end + 1])

		desc_start = mime_end + 2
		description, image_data = split_encoded(data[desc_start:], encoding)
		description = decode_bytestring(description, encoding)
		width, height = get_image_size(image_data)

		return cls(
			type=type_,
			mime_type=mime_type,
			description=description,
			width=width,
			height=height,
			data=image_data,
		)


@attrs(repr=False)
class ID3v2BaseFrame(AttrMapping):
	id = attrib()  # noqa


@attrs(repr=False)
class ID3v2CommentFrame(ID3v2BaseFrame):
	language = attrib()
	description = attrib()
	value = attrib()


@attrs(repr=False)
class ID3v2GenreFrame(ID3v2BaseFrame):
	value = attrib()


@attrs(repr=False)
class ID3v2GEOBFrame(ID3v2BaseFrame):
	mime_type = attrib()
	filename = attrib()
	description = attrib()
	value = attrib()


@attrs(repr=False)
class ID3v2NumberFrame(ID3v2BaseFrame):
	value = attrib()

	@value.validator
	def validate_value(self, attribute, value):
		if not all(char in [*string.digits, '/'] for char in value):
			raise ValueError(
				"Number frame values must consist only of digits and '/'.",
			)

	@property
	def number(self):
		return self.value.split('/')[0]

	@property
	def total(self):
		try:
			tot = self.value.split('/')[1]
		except IndexError:
			tot = None

		return tot


@attrs(repr=False)
class ID3v2NumericTextFrame(ID3v2BaseFrame):
	value = attrib()

	@value.validator
	def validate_value(self, attribute, value):
		if not all(v.isdigit() for v in value):
			raise ValueError("Numeric text frame values must consist only of digits.")


@attrs(repr=False)
class ID3v2PictureFrame(ID3v2BaseFrame):
	value = attrib(converter=ID3v2Picture.load)


@attrs(repr=False)
class ID3v2PrivateFrame(ID3v2BaseFrame):
	owner = attrib()
	value = attrib()


@attrs(repr=False)
class ID3v2SynchronizedLyricsFrame(ID3v2BaseFrame):
	language = attrib()
	timestamp_format = attrib()
	description = attrib()
	value = attrib()


@attrs(repr=False)
class ID3v2TextFrame(ID3v2BaseFrame):
	value = attrib()


@attrs(repr=False)
class ID3v2TimestampFrame(ID3v2BaseFrame):
	value = attrib()


@attrs(repr=False)
class ID3v2UnsynchronizedLyricsFrame(ID3v2BaseFrame):
	language = attrib()
	description = attrib()
	value = attrib()


@attrs(repr=False)
class ID3v2URLLinkFrame(ID3v2BaseFrame):
	value = attrib()


@attrs(repr=False)
class ID3v2UserURLLinkFrame(ID3v2BaseFrame):
	description = attrib()
	value = attrib()


@attrs(repr=False)
class ID3v2UserTextFrame(ID3v2BaseFrame):
	description = attrib()
	value = attrib()


@attrs(repr=False)
class ID3v2YearFrame(ID3v2NumericTextFrame):
	value = attrib()

	@value.validator
	def validate_value(self, attribute, value):
		if not all(
			(
				v.isdigit()
				and len(v) == 4
			)
			for v in value
		):
			raise ValueError("Year frame values must be 4-character number strings.")


@attrs(repr=False)
class ID3v2TDATFrame(ID3v2NumericTextFrame):
	value = attrib()

	@value.validator
	def validate_value(self, attribute, value):
		if not all(
			(
				v.isdigit()
				and len(v) == 4
				and int(v[0:2]) in range(1, 32)
				and int(v[2:4]) in range(1, 13)
			)
			for v in value
		):
			raise ValueError(
				"TDAT frame values must be a 4-character number string in the DDMM format.",
			)


@attrs(repr=False)
class ID3v2TIMEFrame(ID3v2NumericTextFrame):
	value = attrib()

	@value.validator
	def validate_value(self, attribute, value):
		if not all(
			(
				v.isdigit()
				and len(v) == 4
				and int(v[0:2]) in range(0, 24)
				and int(v[2:4]) in range(0, 60)
			)
			for v in value
		):
			raise ValueError(
				"TIME frame values must be a 4-character number string in the HHMM format.",
			)


@attrs(repr=False)
class ID3v2Frame(ID3v2BaseFrame):
	value = attrib()

	# TODO:ID3v2.2
	# TODO: BUF, CNT, CRA, CRM, ETC, EQU, IPL, LNK, MCI, MLL, POP, REV,
	# TODO: RVA, STC, UFI

	# TODO: ID3v2.3
	# TODO: AENC, COMR, ENCR, EQUA, ETCO, GRID, IPLS, LINK, MCDI, MLLT, OWNE
	# TODO: PCNT, POPM, POSS, RBUF, RVAD, RVRB, SYTC, UFID, USER

	# TODO: ID3v2.4
	# TODO: ASPI, EQU2, RVA2, SEEK, SIGN, TDEN, TDOR, TDRC, TDRL, TDTG, TIPL
	# TODO: TMCL, TPRO,

	_FRAME_TYPES = {
		# Complex Text Frames
		'COM': ID3v2CommentFrame,
		'GEO': ID3v2GEOBFrame,
		'TXX': ID3v2UserTextFrame,

		'COMM': ID3v2CommentFrame,
		'GEOB': ID3v2GEOBFrame,
		'PRIV': ID3v2PrivateFrame,
		'TXXX': ID3v2UserTextFrame,

		# Genre Frame
		'TCO': ID3v2GenreFrame,

		'TCON': ID3v2GenreFrame,

		# Lyrics Frames
		'SLT': ID3v2SynchronizedLyricsFrame,
		'ULT': ID3v2UnsynchronizedLyricsFrame,

		'SYLT': ID3v2SynchronizedLyricsFrame,
		'USLT': ID3v2UnsynchronizedLyricsFrame,

		# Number Frames
		'TPA': ID3v2NumberFrame,
		'TRK': ID3v2NumberFrame,

		'TPOS': ID3v2NumberFrame,
		'TRCK': ID3v2NumberFrame,

		# Numeric Text Frames
		'TBP': ID3v2NumericTextFrame,
		'TDA': ID3v2TDATFrame,
		'TDY': ID3v2NumericTextFrame,
		'TIM': ID3v2TIMEFrame,
		'TLE': ID3v2NumericTextFrame,
		'TOR': ID3v2YearFrame,
		'TSI': ID3v2NumericTextFrame,
		'TYE': ID3v2YearFrame,

		'TBPM': ID3v2NumericTextFrame,
		'TDAT': ID3v2TDATFrame,
		'TDLY': ID3v2NumericTextFrame,
		'TIME': ID3v2TIMEFrame,
		'TLEN': ID3v2NumericTextFrame,
		'TORY': ID3v2YearFrame,
		'TSIZ': ID3v2NumericTextFrame,
		'TYER': ID3v2YearFrame,

		# Picture Frames
		'PIC': ID3v2PictureFrame,

		'APIC': ID3v2PictureFrame,

		# Text Frames
		'TAL': ID3v2TextFrame,
		'TCM': ID3v2TextFrame,
		'TCR': ID3v2TextFrame,
		'TEN': ID3v2TextFrame,
		'TFT': ID3v2TextFrame,
		'TKE': ID3v2TextFrame,
		'TLA': ID3v2TextFrame,
		'TMT': ID3v2TextFrame,
		'TOA': ID3v2TextFrame,
		'TOF': ID3v2TextFrame,
		'TOL': ID3v2TextFrame,
		'TOT': ID3v2TextFrame,
		'TP1': ID3v2TextFrame,
		'TP2': ID3v2TextFrame,
		'TP3': ID3v2TextFrame,
		'TP4': ID3v2TextFrame,
		'TPB': ID3v2TextFrame,
		'TRC': ID3v2TextFrame,
		'TRD': ID3v2TextFrame,
		'TSS': ID3v2TextFrame,
		'TT1': ID3v2TextFrame,
		'TT2': ID3v2TextFrame,
		'TT3': ID3v2TextFrame,
		'TXT': ID3v2TextFrame,

		'TALB': ID3v2TextFrame,
		'TCMP': ID3v2TextFrame,
		'TCOM': ID3v2TextFrame,
		'TCOP': ID3v2TextFrame,
		'TENC': ID3v2TextFrame,
		'TEXT': ID3v2TextFrame,
		'TFLT': ID3v2TextFrame,
		'TIT1': ID3v2TextFrame,
		'TIT2': ID3v2TextFrame,
		'TIT3': ID3v2TextFrame,
		'TKEY': ID3v2TextFrame,
		'TLAN': ID3v2TextFrame,
		'TMED': ID3v2TextFrame,
		'TMOO': ID3v2TextFrame,
		'TOAL': ID3v2TextFrame,
		'TOFN': ID3v2TextFrame,
		'TOLY': ID3v2TextFrame,
		'TOPE': ID3v2TextFrame,
		'TOWN': ID3v2TextFrame,
		'TPE1': ID3v2TextFrame,
		'TPE2': ID3v2TextFrame,
		'TPE3': ID3v2TextFrame,
		'TPE4': ID3v2TextFrame,
		'TPUB': ID3v2TextFrame,
		'TRDA': ID3v2TextFrame,
		'TRSN': ID3v2TextFrame,
		'TRSO': ID3v2TextFrame,
		'TSO2': ID3v2TextFrame,
		'TSOA': ID3v2TextFrame,
		'TSOC': ID3v2TextFrame,
		'TSOP': ID3v2TextFrame,
		'TSOT': ID3v2TextFrame,
		'TSRC': ID3v2TextFrame,
		'TSSE': ID3v2TextFrame,
		'TSST': ID3v2TextFrame,

		# Timestamp Frames
		# TODO: Proper.
		'TDRC': ID3v2TimestampFrame,
		'TDRL': ID3v2TimestampFrame,

		# URL Link Frames
		'WAF': ID3v2URLLinkFrame,
		'WAR': ID3v2URLLinkFrame,
		'WAS': ID3v2URLLinkFrame,
		'WCM': ID3v2URLLinkFrame,
		'WCP': ID3v2URLLinkFrame,
		'WPB': ID3v2URLLinkFrame,
		'WXX': ID3v2UserURLLinkFrame,

		'WCOM': ID3v2URLLinkFrame,
		'WCOP': ID3v2URLLinkFrame,
		'WOAF': ID3v2URLLinkFrame,
		'WOAR': ID3v2URLLinkFrame,
		'WOAS': ID3v2URLLinkFrame,
		'WORS': ID3v2URLLinkFrame,
		'WPAY': ID3v2URLLinkFrame,
		'WPUB': ID3v2URLLinkFrame,
		'WXXX': ID3v2UserURLLinkFrame,
	}

	@datareader
	@classmethod
	def load(cls, data, struct_pattern, size_len, per_byte):
		try:
			frame = struct.unpack(struct_pattern, data.read(struct.calcsize(struct_pattern)))
		except struct.error:
			raise InvalidFrame("Not enough data.")

		frame_size = decode_synchsafe_int(frame[1:1 + size_len], per_byte)
		if frame_size == 0:
			raise InvalidFrame("Not a valid ID3v2 frame")

		frame_id = frame[0].decode('iso-8859-1')
		frame_type = ID3v2Frame._FRAME_TYPES.get(frame_id, cls)
		frame_data = data.read(frame_size)

		# TODO: Move logic into frame classes?
		args = [frame_id]
		if frame_type is ID3v2CommentFrame:
			encoding = determine_encoding(frame_data[0:1])

			language = decode_bytestring(frame_data[1:4])
			args.append(language)

			values = split_encoded(frame_data[4:], encoding)

			# Ignore empty comments.
			if len(values) < 2:
				return None

			args.extend(
				decode_bytestring(v, encoding)
				for v in values
			)
		elif frame_type is ID3v2GenreFrame:
			encoding = determine_encoding(frame_data[0:1])

			remainder = frame_data[1:]
			values = []
			while True:
				split = split_encoded(remainder, encoding)
				values.extend(
					decode_bytestring(v, encoding)
					for v in split
				)

				if len(split) < 2:
					break

				remainder = split[1]

			genres = []
			for value in values:
				match = _genre_re.match(value)

				if match['name']:
					genres.append(match['name'])
				elif match['id']:
					if match['id'].isdigit() and int(match['id']):
						try:
							genres.append(ID3v1Genres[int(match['id'])])
						except IndexError:
							genres.append(value)
					elif match['id'] == 'CR':
						genres.append('Cover')
					elif match['id'] == 'RX':
						genres.append('Remix')

			args.append(genres)
		elif frame_type is ID3v2GEOBFrame:
			encoding = determine_encoding(frame_data[0:1])

			mime_type, remainder = split_encoded(frame_data[1:], encoding)
			filename, remainder = split_encoded(remainder, encoding)
			description, value = split_encoded(remainder, encoding)

			values = [decode_bytestring(mime_type)]
			values.extend(
				decode_bytestring(v, encoding)
				for v in [filename, description]
			)
			values.append(value)

			args.extend(values)
		elif frame_type is ID3v2PictureFrame:
			args.append(frame_data)
		elif frame_type is ID3v2PrivateFrame:
			owner_end = frame_data.index(b'\x00')
			args.append(frame_data[0:owner_end].decode('iso-8859-1'))
			args.append(frame_data[owner_end + 1:])
		elif frame_type is ID3v2UnsynchronizedLyricsFrame:
			encoding = determine_encoding(frame_data[0:1])

			language = decode_bytestring(frame_data[1:4])
			args.append(language)

			for v in split_encoded(frame_data[4:], encoding):
				args.append(decode_bytestring(v, encoding))
		elif frame_type is ID3v2URLLinkFrame:
			args.append(unquote(decode_bytestring(frame_data)))
		elif frame_type is ID3v2UserURLLinkFrame:
			encoding = determine_encoding(frame_data)

			description, url = split_encoded(frame_data[1:], encoding)
			args.append(decode_bytestring(description, encoding))
			args.append(unquote(decode_bytestring(url)))
		elif issubclass(
			frame_type,
			(
				ID3v2NumberFrame,
				ID3v2UserTextFrame,
			),
		):
			encoding = determine_encoding(frame_data[0:1])
			args.append(decode_bytestring(frame_data[1:], encoding))
		elif issubclass(
			frame_type,
			(
				ID3v2NumericTextFrame,
				ID3v2TextFrame,
				ID3v2TimestampFrame,
			),
		):
			encoding = determine_encoding(frame_data[0:1])
			values = [
				decode_bytestring(value, encoding)
				for value in split_encoded(frame_data[1:], encoding)
				if value
			]
			args.append(values)
		elif frame_type is ID3v2Frame:
			args.append(frame_data)
		else:
			args.append(decode_bytestring(frame_data))

		try:
			return frame_type(*args)
		except (TypeError, ValueError):  # Bad frame value.
			return None
